﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using Castle.Core.Logging;
using StudioX.Collections.Extensions;
using StudioX.Configuration.Startup;
using StudioX.Dependency;

namespace StudioX.Modules
{
    /// <summary>
    ///     This class must be implemented by all module definition classes.
    /// </summary>
    /// <remarks>
    ///     A module definition class is generally located in it's own assembly
    ///     and implements some action in module events on application startup and shutdown.
    ///     It also defines depended modules.
    /// </remarks>
    public abstract class StudioXModule
    {
        /// <summary>
        ///     Gets a reference to the IOC manager.
        /// </summary>
        protected internal IIocManager IocManager { get; internal set; }

        /// <summary>
        ///     Gets a reference to the StudioX configuration.
        /// </summary>
        protected internal IStudioXStartupConfiguration Configuration { get; internal set; }

        /// <summary>
        ///     Gets or sets the logger.
        /// </summary>
        public ILogger Logger { get; set; }

        protected StudioXModule()
        {
            Logger = NullLogger.Instance;
        }

        /// <summary>
        ///     This is the first event called on application startup.
        ///     Codes can be placed here to run before dependency injection registrations.
        /// </summary>
        public virtual void PreInitialize()
        {
        }

        /// <summary>
        ///     This method is used to register dependencies for this module.
        /// </summary>
        public virtual void Initialize()
        {
        }

        /// <summary>
        ///     This method is called lastly on application startup.
        /// </summary>
        public virtual void PostInitialize()
        {
        }

        /// <summary>
        ///     This method is called when the application is being shutdown.
        /// </summary>
        public virtual void Shutdown()
        {
        }

        public virtual Assembly[] GetAdditionalAssemblies()
        {
            return new Assembly[0];
        }

        /// <summary>
        ///     Checks if given type is an StudioX module class.
        /// </summary>
        /// <param name="type">Type to check</param>
        public static bool IsStudioXModule(Type type)
        {
            var typeInfo = type.GetTypeInfo();
            return
                typeInfo.IsClass &&
                !typeInfo.IsAbstract &&
                !typeInfo.IsGenericType &&
                typeof(StudioXModule).IsAssignableFrom(type);
        }

        /// <summary>
        ///     Finds direct depended modules of a module (excluding given module).
        /// </summary>
        public static List<Type> FindDependedModuleTypes(Type moduleType)
        {
            if (!IsStudioXModule(moduleType))
            {
                throw new StudioXInitializationException("This type is not an StudioX module: " +
                                                         moduleType.AssemblyQualifiedName);
            }

            var list = new List<Type>();

            if (moduleType.GetTypeInfo().IsDefined(typeof(DependsOnAttribute), true))
            {
                var dependsOnAttributes =
                    moduleType.GetTypeInfo()
                        .GetCustomAttributes(typeof(DependsOnAttribute), true)
                        .Cast<DependsOnAttribute>();
                foreach (var dependsOnAttribute in dependsOnAttributes)
                {
                    foreach (var dependedModuleType in dependsOnAttribute.DependedModuleTypes)
                    {
                        list.Add(dependedModuleType);
                    }
                }
            }

            return list;
        }

        public static List<Type> FindDependedModuleTypesRecursivelyIncludingGivenModule(Type moduleType)
        {
            var list = new List<Type>();
            AddModuleAndDependenciesRecursively(list, moduleType);
            list.AddIfNotContains(typeof(StudioXKernelModule));
            return list;
        }

        private static void AddModuleAndDependenciesRecursively(List<Type> modules, Type module)
        {
            if (!IsStudioXModule(module))
            {
                throw new StudioXInitializationException("This type is not an StudioX module: " +
                                                         module.AssemblyQualifiedName);
            }

            if (modules.Contains(module))
            {
                return;
            }

            modules.Add(module);

            var dependedModules = FindDependedModuleTypes(module);
            foreach (var dependedModule in dependedModules)
            {
                AddModuleAndDependenciesRecursively(modules, dependedModule);
            }
        }
    }
}